@@ -1,6 +1,7 @@
 from __future__ import annotations
 
-from typing import Any
+import time
+from typing import Any, Callable
 
 from linkplay.consts import (
     INPUT_MODE_MAP,
@@ -20,7 +21,13 @@ from linkplay.consts import (
     SpeakerType,
 )
 from linkplay.endpoint import LinkPlayEndpoint
-from linkplay.utils import fixup_player_properties
+from linkplay.exceptions import LinkPlayInvalidDataException
+from linkplay.manufacturers import MANUFACTURER_WIIM, get_info_from_project
+from linkplay.utils import (
+    equalizer_mode_from_number_mapping,
+    equalizer_mode_to_number_mapping,
+    fixup_player_properties,
+)
 
 
 class LinkPlayDevice:
@@ -29,6 +36,8 @@ class LinkPlayDevice:
     bridge: LinkPlayBridge
     properties: dict[DeviceAttribute, str]
 
+    controller: Callable[[], None] | None = None
+
     def __init__(self, bridge: LinkPlayBridge):
         self.bridge = bridge
         self.properties = dict.fromkeys(DeviceAttribute.__members__.values(), "")
@@ -37,6 +46,10 @@ class LinkPlayDevice:
         """Return the state of the LinkPlayDevice."""
         return {"properties": self.properties}
 
+    def set_callback(self, controller: Callable[[], None]) -> None:
+        """Sets a callback function to notify events."""
+        self.controller = controller
+
     async def update_status(self) -> None:
         """Update the device status."""
         self.properties = await self.bridge.json_request(LinkPlayCommand.DEVICE_STATUS)  # type: ignore[assignment]
@@ -55,6 +68,22 @@ class LinkPlayDevice:
         """The name of the device."""
         return self.properties.get(DeviceAttribute.DEVICE_NAME, "")
 
+    @property
+    def manufacturer(self) -> str:
+        """The manufacturer of the device."""
+        manufacturer, _ = get_info_from_project(
+            self.properties.get(DeviceAttribute.PROJECT, "")
+        )
+        return manufacturer
+
+    @property
+    def model(self) -> str:
+        """The model of the device."""
+        _, model = get_info_from_project(
+            self.properties.get(DeviceAttribute.PROJECT, "")
+        )
+        return model
+
     @property
     def playmode_support(self) -> list[PlayingMode]:
         """Returns the player playmode support."""
@@ -64,27 +93,48 @@ class LinkPlayDevice:
         )
 
         playing_modes = [INPUT_MODE_MAP[flag] for flag in flags]
-        playing_modes.insert(0, PlayingMode.NETWORK) # always supported
+        playing_modes.insert(0, PlayingMode.NETWORK)  # always supported
         return playing_modes
 
     @property
-    def eth(self) -> str:
+    def mac(self) -> str | None:
+        """Returns the mac address."""
+        mac = self.properties.get(DeviceAttribute.ETH_MAC_ADDRESS)
+        if mac == "00:00:00:00:00:00" or mac is None:
+            mac = self.properties.get(DeviceAttribute.STA_MAC_ADDRESS)
+        if mac == "00:00:00:00:00:00" or mac is None:
+            mac = self.properties.get(DeviceAttribute.MAC_ADDRESS)
+        return mac
+
+    @property
+    def eth(self) -> str | None:
         """Returns the ethernet address."""
-        return (
-            self.properties[DeviceAttribute.ETH2]
-            if DeviceAttribute.ETH_DHCP in self.properties
-            else self.properties[DeviceAttribute.APCLI0]
-        )
+        eth = self.properties.get(DeviceAttribute.ETH2)
+        if eth == "0.0.0.0" or eth == "" or eth is None:
+            eth = self.properties.get(DeviceAttribute.ETH0)
+        if eth == "0.0.0.0" or eth == "" or eth is None:
+            eth = self.properties.get(DeviceAttribute.APCLI0)
+        return eth
+
+    async def timesync(self) -> None:
+        """Sync the time."""
+        timestamp = time.strftime("%Y%m%d%H%M%S")
+        await self.bridge.request(LinkPlayCommand.TIMESYNC.format(timestamp))
+
 
 class LinkPlayPlayer:
     """Represents a LinkPlay player."""
 
     bridge: LinkPlayBridge
     properties: dict[PlayerAttribute, str]
+    custom_properties: dict[PlayerAttribute, str]
+
+    previous_playing_mode: PlayingMode | None = None
 
     def __init__(self, bridge: LinkPlayBridge):
         self.bridge = bridge
         self.properties = dict.fromkeys(PlayerAttribute.__members__.values(), "")
+        self.custom_properties = dict.fromkeys(PlayerAttribute.__members__.values(), "")
 
     def to_dict(self):
         """Return the state of the LinkPlayPlayer."""
@@ -98,6 +148,20 @@ class LinkPlayPlayer:
 
         self.properties = fixup_player_properties(properties)
 
+        # handle multiroom changes
+        if self.bridge.device.controller is not None and (
+            (
+                self.previous_playing_mode != PlayingMode.FOLLOWER
+                and self.play_mode == PlayingMode.FOLLOWER
+            )
+            or (
+                self.previous_playing_mode == PlayingMode.FOLLOWER
+                and self.play_mode != PlayingMode.FOLLOWER
+            )
+        ):
+            self.bridge.device.controller()
+        self.previous_playing_mode = self.play_mode
+
     async def next(self) -> None:
         """Play the next song in the playlist."""
         await self.bridge.request(LinkPlayCommand.NEXT)
@@ -151,7 +215,37 @@ class LinkPlayPlayer:
 
     async def set_equalizer_mode(self, mode: EqualizerMode) -> None:
         """Set the equalizer mode."""
-        await self.bridge.request(LinkPlayCommand.EQUALIZER_MODE.format(mode))  # type: ignore[str-format]
+        if self.bridge.device.manufacturer == MANUFACTURER_WIIM:
+            await self._set_wiim_equalizer_mode(mode)
+        else:
+            await self._set_normal_equalizer_mode(mode)
+
+    async def _set_wiim_equalizer_mode(self, mode: EqualizerMode) -> None:
+        # WiiM devices have a different equalizer mode handling
+        # and don't support the general equalizer mode command
+        if mode not in self.available_equalizer_modes:
+            raise ValueError(
+                f"Invalid equalizer mode {mode}. Allowed equalizer modes are {self.available_equalizer_modes}."
+            )
+
+        if mode == EqualizerMode.NONE:
+            await self.bridge.json_request(LinkPlayCommand.WIIM_EQUALIZER_OFF)
+        else:
+            await self.bridge.json_request(LinkPlayCommand.WIIM_EQ_LOAD.format(mode))
+        # WiiM doesn't update the property after setting it
+        self.custom_properties[PlayerAttribute.EQUALIZER_MODE] = mode
+
+    async def _set_normal_equalizer_mode(self, mode: EqualizerMode) -> None:
+        """Set the equalizer mode."""
+        equalizer_mode_as_number = equalizer_mode_to_number_mapping(mode)
+        if equalizer_mode_as_number is None:
+            raise ValueError(
+                f"Invalid equalizer mode {mode}. Allowed equalizer modes are {self.available_equalizer_modes}."
+            )
+
+        await self.bridge.request(
+            LinkPlayCommand.EQUALIZER_MODE.format(equalizer_mode_as_number)
+        )
 
     async def set_loop_mode(self, mode: LoopMode) -> None:
         """Set the loop mode."""
@@ -174,11 +268,14 @@ class LinkPlayPlayer:
             )
         await self.bridge.request(LinkPlayCommand.PLAY_PRESET.format(preset_number))
 
-    async def timesync(self) -> None:
-        """Sync the time."""
-        import time
-        timestamp = time.strftime('%Y%m%d%H%M%S')
-        await self.bridge.request(LinkPlayCommand.TIMESYNC.format(timestamp))
+    async def seek(self, position: int) -> None:
+        """Seek to a position."""
+        if (
+            self.total_length_in_seconds > 0
+            and position >= 0
+            and position <= self.total_length_in_seconds
+        ):
+            await self.bridge.request(LinkPlayCommand.SEEK.format(position))
 
     @property
     def muted(self) -> bool:
@@ -218,6 +315,16 @@ class LinkPlayPlayer:
         """Returns the total length of the track in milliseconds."""
         return int(self.properties.get(PlayerAttribute.TOTAL_LENGTH, 0))
 
+    @property
+    def current_position_in_seconds(self) -> int:
+        """Returns the current position of the track in seconds."""
+        return int(int(self.properties.get(PlayerAttribute.CURRENT_POSITION, 0)) / 1000)
+
+    @property
+    def total_length_in_seconds(self) -> int:
+        """Returns the total length of the track in seconds."""
+        return int(int(self.properties.get(PlayerAttribute.TOTAL_LENGTH, 0)) / 1000)
+
     @property
     def status(self) -> PlayingStatus:
         """Returns the current playing status."""
@@ -228,9 +335,62 @@ class LinkPlayPlayer:
     @property
     def equalizer_mode(self) -> EqualizerMode:
         """Returns the current equalizer mode."""
-        return EqualizerMode(
-            self.properties.get(PlayerAttribute.EQUALIZER_MODE, EqualizerMode.CLASSIC)
-        )
+        try:
+            if self.bridge.device.manufacturer == MANUFACTURER_WIIM:
+                # WiiM devices have a different equalizer mode handling
+                # and will never ever return what equalizer mode they are in
+                return EqualizerMode(
+                    self.custom_properties.get(
+                        PlayerAttribute.EQUALIZER_MODE, EqualizerMode.NONE
+                    )
+                )
+
+            return EqualizerMode(
+                equalizer_mode_from_number_mapping(
+                    self.properties.get(PlayerAttribute.EQUALIZER_MODE)
+                )
+                or EqualizerMode.NONE
+            )
+        except ValueError:
+            return EqualizerMode.NONE
+
+    @property
+    def available_equalizer_modes(self) -> list[EqualizerMode]:
+        """Returns the available equalizer modes."""
+        if self.bridge.device.manufacturer == MANUFACTURER_WIIM:
+            return [
+                EqualizerMode.NONE,
+                EqualizerMode.FLAT,
+                EqualizerMode.ACOUSTIC,
+                EqualizerMode.BASS_BOOSTER,
+                EqualizerMode.BASS_REDUCER,
+                EqualizerMode.CLASSICAL,
+                EqualizerMode.DANCE,
+                EqualizerMode.DEEP,
+                EqualizerMode.ELECTRONIC,
+                EqualizerMode.HIP_HOP,
+                EqualizerMode.JAZZ,
+                EqualizerMode.LATIN,
+                EqualizerMode.LOUDNESS,
+                EqualizerMode.LOUNGE,
+                EqualizerMode.PIANO,
+                EqualizerMode.POP,
+                EqualizerMode.R_B,
+                EqualizerMode.ROCK,
+                EqualizerMode.SMALL_SPEAKERS,
+                EqualizerMode.SPOKEN_WORD,
+                EqualizerMode.TREBLE_BOOSTER,
+                EqualizerMode.TREBLE_REDUCER,
+                EqualizerMode.VOCAL_BOOSTER,
+            ]
+
+        return [
+            EqualizerMode.NONE,
+            EqualizerMode.CLASSIC,
+            EqualizerMode.POP,
+            EqualizerMode.JAZZ,
+            EqualizerMode.VOCAL,
+        ]
 
     @property
     def speaker_type(self) -> SpeakerType:
@@ -329,26 +489,31 @@ class LinkPlayMultiroom:
 
     async def update_status(self, bridges: list[LinkPlayBridge]) -> None:
         """Updates the multiroom status."""
-        properties: dict[Any, Any] = await self.leader.json_request(
-            LinkPlayCommand.MULTIROOM_LIST
-        )
+        try:
+            properties: dict[Any, Any] = await self.leader.json_request(
+                LinkPlayCommand.MULTIROOM_LIST
+            )
 
-        self.followers = []
-        if int(properties[MultiroomAttribute.NUM_FOLLOWERS]) == 0:
-            return
+            self.followers = []
+            if int(properties[MultiroomAttribute.NUM_FOLLOWERS]) == 0:
+                return
 
-        follower_uuids = [
-            follower[MultiroomAttribute.UUID]
-            for follower in properties[MultiroomAttribute.FOLLOWER_LIST]
-        ]
-        new_followers = [
-            bridge for bridge in bridges if bridge.device.uuid in follower_uuids
-        ]
-        self.followers.extend(new_followers)
+            follower_uuids = [
+                follower[MultiroomAttribute.UUID]
+                for follower in properties[MultiroomAttribute.FOLLOWER_LIST]
+            ]
+            new_followers = [
+                bridge for bridge in bridges if bridge.device.uuid in follower_uuids
+            ]
+            self.followers.extend(new_followers)
+        except LinkPlayInvalidDataException as exc:
+            LOGGER.exception(exc)
 
     async def ungroup(self) -> None:
         """Ungroups the multiroom group."""
         await self.leader.request(LinkPlayCommand.MULTIROOM_UNGROUP)
+        for follewer in self.followers:
+            follewer.multiroom = None
         self.followers = []
 
     async def add_follower(self, follower: LinkPlayBridge) -> None:
@@ -356,14 +521,18 @@ class LinkPlayMultiroom:
         await follower.request(
             LinkPlayCommand.MULTIROOM_JOIN.format(self.leader.device.eth)
         )  # type: ignore[str-format]
-        self.followers.append(follower)
+        if follower not in self.followers:
+            follower.multiroom = self
+            self.followers.append(follower)
 
     async def remove_follower(self, follower: LinkPlayBridge) -> None:
         """Removes a follower from the multiroom group."""
         await self.leader.request(
             LinkPlayCommand.MULTIROOM_KICK.format(follower.device.eth)
         )  # type: ignore[str-format]
-        self.followers.remove(follower)
+        if follower in self.followers:
+            follower.multiroom = None
+            self.followers.remove(follower)
 
     async def set_volume(self, value: int) -> None:
         """Sets the volume for the multiroom group."""
